@@ -25,8 +25,8 @@ var Guacamole = Guacamole || {};
  * provided tunnel, updating its display using one or more canvas elements.
  * 
  * @constructor
- * @param {Guacamole.Tunnel} tunnel The tunnel to use to send and receive
- *                                  Guacamole instructions.
+ * @param {!Guacamole.Tunnel} tunnel
+ *     The tunnel to use to send and receive Guacamole instructions.
  */
 Guacamole.Client = function(tunnel) {
 
@@ -46,7 +46,9 @@ Guacamole.Client = function(tunnel) {
 
     /**
      * Translation from Guacamole protocol line caps to Layer line caps.
+     *
      * @private
+     * @type {!Object.<number, string>}
      */
     var lineCap = {
         0: "butt",
@@ -56,7 +58,9 @@ Guacamole.Client = function(tunnel) {
 
     /**
      * Translation from Guacamole protocol line caps to Layer line caps.
+     *
      * @private
+     * @type {!Object.<number, string>}
      */
     var lineJoin = {
         0: "bevel",
@@ -68,7 +72,7 @@ Guacamole.Client = function(tunnel) {
      * The underlying Guacamole display.
      *
      * @private
-     * @type {Guacamole.Display}
+     * @type {!Guacamole.Display}
      */
     var display = new Guacamole.Display();
 
@@ -76,7 +80,7 @@ Guacamole.Client = function(tunnel) {
      * All available layers and buffers
      *
      * @private
-     * @type {Object.<Number, (Guacamole.Display.VisibleLayer|Guacamole.Layer)>}
+     * @type {!Object.<number, (Guacamole.Display.VisibleLayer|Guacamole.Layer)>}
      */
     var layers = {};
     
@@ -85,7 +89,7 @@ Guacamole.Client = function(tunnel) {
      * be empty, but audio players may be allocated by the server upon request.
      *
      * @private
-     * @type {Object.<Number, Guacamole.AudioPlayer>}
+     * @type {!Object.<number, Guacamole.AudioPlayer>}
      */
     var audioPlayers = {};
 
@@ -94,7 +98,7 @@ Guacamole.Client = function(tunnel) {
      * be empty, but video players may be allocated by the server upon request.
      *
      * @private
-     * @type {Object.<Number, Guacamole.VideoPlayer>}
+     * @type {!Object.<number, Guacamole.VideoPlayer>}
      */
     var videoPlayers = {};
 
@@ -109,7 +113,7 @@ Guacamole.Client = function(tunnel) {
      * Guacamole server.
      *
      * @private
-     * @type {Guacamole.Object[]}
+     * @type {!Guacamole.Object[]}
      */
     var objects = [];
 
@@ -138,7 +142,7 @@ Guacamole.Client = function(tunnel) {
      * effectively an independent, compressed snapshot of protocol and display
      * state. Invoking this function implicitly flushes the display.
      *
-     * @param {function} callback
+     * @param {!function} callback
      *     Callback which should be invoked once the state object is ready. The
      *     state object will be passed to the callback as the sole parameter.
      *     This callback may be invoked immediately, or later as the display
@@ -207,7 +211,7 @@ Guacamole.Client = function(tunnel) {
      * object from a prior call to exportState(). The Guacamole.Client instance
      * used to export that state need not be the same as this instance.
      *
-     * @param {Object} state
+     * @param {!object} state
      *     An opaque representation of Guacamole.Client state from a prior call
      *     to exportState().
      *
@@ -224,11 +228,14 @@ Guacamole.Client = function(tunnel) {
         currentState = state.currentState;
         currentTimestamp = state.currentTimestamp;
 
+        // Cancel any pending display operations/frames
+        display.cancel();
+
         // Dispose of all layers
         for (key in layers) {
             index = parseInt(key);
             if (index > 0)
-                display.dispose(layers[key]);
+                layers[key].dispose();
         }
 
         layers = {};
@@ -280,8 +287,8 @@ Guacamole.Client = function(tunnel) {
      * contains an Element which can be added to the DOM, causing the
      * display to become visible.
      * 
-     * @return {Guacamole.Display} The underlying display of this
-     *                             Guacamole.Client.
+     * @return {!Guacamole.Display}
+     *     The underlying display of this Guacamole.Client.
      */
     this.getDisplay = function() {
         return display;
@@ -290,8 +297,11 @@ Guacamole.Client = function(tunnel) {
     /**
      * Sends the current size of the screen.
      * 
-     * @param {Number} width The width of the screen.
-     * @param {Number} height The height of the screen.
+     * @param {!number} width
+     *     The width of the screen.
+     *
+     * @param {!number} height
+     *     The height of the screen.
      */
     this.sendSize = function(width, height) {
 
@@ -307,9 +317,11 @@ Guacamole.Client = function(tunnel) {
      * Sends a key event having the given properties as if the user
      * pressed or released a key.
      * 
-     * @param {Boolean} pressed Whether the key is pressed (true) or released
-     *                          (false).
-     * @param {Number} keysym The keysym of the key being pressed or released.
+     * @param {!boolean} pressed
+     *     Whether the key is pressed (true) or released (false).
+     *
+     * @param {!number} keysym
+     *     The keysym of the key being pressed or released.
      */
     this.sendKeyEvent = function(pressed, keysym) {
         // Do not send requests if not connected
@@ -323,19 +335,33 @@ Guacamole.Client = function(tunnel) {
      * Sends a mouse event having the properties provided by the given mouse
      * state.
      * 
-     * @param {Guacamole.Mouse.State} mouseState The state of the mouse to send
-     *                                           in the mouse event.
+     * @param {!Guacamole.Mouse.State} mouseState
+     *     The state of the mouse to send in the mouse event.
+     *
+     * @param {boolean} [applyDisplayScale=false]
+     *     Whether the provided mouse state uses local display units, rather
+     *     than remote display units, and should be scaled to match the
+     *     {@link Guacamole.Display}.
      */
-    this.sendMouseState = function(mouseState) {
+    this.sendMouseState = function sendMouseState(mouseState, applyDisplayScale) {
 
         // Do not send requests if not connected
         if (!isConnected())
             return;
 
+        var x = mouseState.x;
+        var y = mouseState.y;
+
+        // Translate for display units if requested
+        if (applyDisplayScale) {
+            x /= display.getScale();
+            y /= display.getScale();
+        }
+
         // Update client-side cursor
         display.moveCursor(
-            Math.floor(mouseState.x),
-            Math.floor(mouseState.y)
+            Math.floor(x),
+            Math.floor(y)
         );
 
         // Build mask
@@ -347,7 +373,40 @@ Guacamole.Client = function(tunnel) {
         if (mouseState.down)   buttonMask |= 16;
 
         // Send message
-        tunnel.sendMessage("mouse", Math.floor(mouseState.x), Math.floor(mouseState.y), buttonMask);
+        tunnel.sendMessage("mouse", Math.floor(x), Math.floor(y), buttonMask);
+    };
+
+    /**
+     * Sends a touch event having the properties provided by the given touch
+     * state.
+     *
+     * @param {!Guacamole.Touch.State} touchState
+     *     The state of the touch contact to send in the touch event.
+     *
+     * @param {boolean} [applyDisplayScale=false]
+     *     Whether the provided touch state uses local display units, rather
+     *     than remote display units, and should be scaled to match the
+     *     {@link Guacamole.Display}.
+     */
+    this.sendTouchState = function sendTouchState(touchState, applyDisplayScale) {
+
+        // Do not send requests if not connected
+        if (!isConnected())
+            return;
+
+        var x = touchState.x;
+        var y = touchState.y;
+
+        // Translate for display units if requested
+        if (applyDisplayScale) {
+            x /= display.getScale();
+            y /= display.getScale();
+        }
+
+        tunnel.sendMessage('touch', touchState.id, Math.floor(x), Math.floor(y),
+            Math.floor(touchState.radiusX), Math.floor(touchState.radiusY),
+            touchState.angle, touchState.force);
+
     };
 
     /**
@@ -359,7 +418,7 @@ Guacamole.Client = function(tunnel) {
      * instruction which creates the stream, such as a "clipboard", "file", or
      * "pipe" instruction.
      *
-     * @returns {Guacamole.OutputStream}
+     * @returns {!Guacamole.OutputStream}
      *     A new Guacamole.OutputStream with a newly-allocated index and
      *     associated with this Guacamole.Client.
      */
@@ -379,11 +438,11 @@ Guacamole.Client = function(tunnel) {
      * mimetype will be sent along the returned stream. The instruction
      * necessary to create this stream will automatically be sent.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the audio data that will be sent along the returned
      *     stream.
      *
-     * @return {Guacamole.OutputStream}
+     * @return {!Guacamole.OutputStream}
      *     The created audio stream.
      */
     this.createAudioStream = function(mimetype) {
@@ -400,9 +459,14 @@ Guacamole.Client = function(tunnel) {
      * filename. The instruction necessary to create this stream will
      * automatically be sent.
      *
-     * @param {String} mimetype The mimetype of the file being sent.
-     * @param {String} filename The filename of the file being sent.
-     * @return {Guacamole.OutputStream} The created file stream.
+     * @param {!string} mimetype
+     *     The mimetype of the file being sent.
+     *
+     * @param {!string} filename
+     *     The filename of the file being sent.
+     *
+     * @return {!Guacamole.OutputStream}
+     *     The created file stream.
      */
     this.createFileStream = function(mimetype, filename) {
 
@@ -417,9 +481,14 @@ Guacamole.Client = function(tunnel) {
      * Opens a new pipe for writing, having the given name and mimetype. The
      * instruction necessary to create this stream will automatically be sent.
      *
-     * @param {String} mimetype The mimetype of the data being sent.
-     * @param {String} name The name of the pipe.
-     * @return {Guacamole.OutputStream} The created file stream.
+     * @param {!string} mimetype
+     *     The mimetype of the data being sent.
+     *
+     * @param {!string} name
+     *     The name of the pipe.
+     *
+     * @return {!Guacamole.OutputStream}
+     *     The created file stream.
      */
     this.createPipeStream = function(mimetype, name) {
 
@@ -434,9 +503,14 @@ Guacamole.Client = function(tunnel) {
      * Opens a new clipboard object for writing, having the given mimetype. The
      * instruction necessary to create this stream will automatically be sent.
      *
-     * @param {String} mimetype The mimetype of the data being sent.
-     * @param {String} name The name of the pipe.
-     * @return {Guacamole.OutputStream} The created file stream.
+     * @param {!string} mimetype
+     *     The mimetype of the data being sent.
+     *
+     * @param {!string} name
+     *     The name of the pipe.
+     *
+     * @return {!Guacamole.OutputStream}
+     *     The created file stream.
      */
     this.createClipboardStream = function(mimetype) {
 
@@ -454,13 +528,13 @@ Guacamole.Client = function(tunnel) {
      * of the following stream. The instruction necessary to create this stream
      * will automatically be sent.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the data being sent.
      *
-     * @param {String} name
+     * @param {!string} name
      *     The name of the connection parameter to attempt to update.
      *
-     * @return {Guacamole.OutputStream}
+     * @return {!Guacamole.OutputStream}
      *     The created argument value stream.
      */
     this.createArgumentValueStream = function createArgumentValueStream(mimetype, name) {
@@ -478,23 +552,23 @@ Guacamole.Client = function(tunnel) {
      * dictated by the object itself. The instruction necessary to create this
      * stream will automatically be sent.
      *
-     * @param {Number} index
+     * @param {!number} index
      *     The index of the object for which the output stream is being
      *     created.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the data which will be sent to the output stream.
      *
-     * @param {String} name
+     * @param {!string} name
      *     The defined name of an output stream within the given object.
      *
-     * @returns {Guacamole.OutputStream}
+     * @returns {!Guacamole.OutputStream}
      *     An output stream which will write blobs to the named output stream
      *     of the given object.
      */
     this.createObjectOutputStream = function createObjectOutputStream(index, mimetype, name) {
 
-        // Allocate and ssociate stream with object metadata
+        // Allocate and associate stream with object metadata
         var stream = guac_client.createOutputStream();
         tunnel.sendMessage("put", index, stream.index, mimetype, name);
         return stream;
@@ -505,11 +579,11 @@ Guacamole.Client = function(tunnel) {
      * Requests read access to the input stream having the given name. If
      * successful, a new input stream will be created.
      *
-     * @param {Number} index
+     * @param {!number} index
      *     The index of the object from which the input stream is being
      *     requested.
      *
-     * @param {String} name
+     * @param {!string} name
      *     The name of the input stream to request.
      */
     this.requestObjectInputStream = function requestObjectInputStream(index, name) {
@@ -524,11 +598,14 @@ Guacamole.Client = function(tunnel) {
     /**
      * Acknowledge receipt of a blob on the stream with the given index.
      * 
-     * @param {Number} index The index of the stream associated with the
-     *                       received blob.
-     * @param {String} message A human-readable message describing the error
-     *                         or status.
-     * @param {Number} code The error code, if any, or 0 for success.
+     * @param {!number} index
+     *     The index of the stream associated with the received blob.
+     *
+     * @param {!string} message
+     *     A human-readable message describing the error or status.
+     *
+     * @param {!number} code
+     *     The error code, if any, or 0 for success.
      */
     this.sendAck = function(index, message, code) {
 
@@ -542,8 +619,11 @@ Guacamole.Client = function(tunnel) {
     /**
      * Given the index of a file, writes a blob of data to that file.
      * 
-     * @param {Number} index The index of the file to write to.
-     * @param {String} data Base64-encoded data to write to the file.
+     * @param {!number} index
+     *     The index of the file to write to.
+     *
+     * @param {!string} data
+     *     Base64-encoded data to write to the file.
      */
     this.sendBlob = function(index, data) {
 
@@ -560,7 +640,7 @@ Guacamole.Client = function(tunnel) {
      * stream is closed, and the index will be made available for reuse in
      * future streams.
      * 
-     * @param {Number} index
+     * @param {!number} index
      *     The index of the stream to end.
      */
     this.endStream = function(index) {
@@ -584,7 +664,8 @@ Guacamole.Client = function(tunnel) {
      * Fired whenever the state of this Guacamole.Client changes.
      * 
      * @event
-     * @param {Number} state The new state of the client.
+     * @param {!number} state
+     *     The new state of the client.
      */
     this.onstatechange = null;
 
@@ -592,7 +673,8 @@ Guacamole.Client = function(tunnel) {
      * Fired when the remote client sends a name update.
      * 
      * @event
-     * @param {String} name The new name of this client.
+     * @param {!string} name
+     *     The new name of this client.
      */
     this.onname = null;
 
@@ -601,25 +683,83 @@ Guacamole.Client = function(tunnel) {
      * is being closed.
      * 
      * @event
-     * @param {Guacamole.Status} status A status object which describes the
-     *                                  error.
+     * @param {!Guacamole.Status} status
+     *     A status object which describes the error.
      */
     this.onerror = null;
+    
+    /**
+     * Fired when an arbitrary message is received from the tunnel that should
+     * be processed by the client. By default, additional message-specific
+     * events such as "onjoin" and "onleave" will fire for the received message
+     * after this event has been processed. An event handler for "onmsg" need
+     * not be supplied if "onjoin" and/or "onleave" will be used.
+     * 
+     * @event
+     * @param {!number} msgcode
+     *     A status code sent by the remote server that indicates the nature of
+     *     the message that is being sent to the client.
+     *     
+     * @param {string[]} args
+     *     An array of arguments to be processed with the message sent to the
+     *     client.
+     *
+     * @return {boolean}
+     *     true if message-specific events such as "onjoin" and
+     *     "onleave" should be fired for this message, false otherwise. If
+     *     no value is returned, message-specific events will be allowed to
+     *     fire.
+     */
+    this.onmsg = null;
+
+    /**
+     * Fired when a user joins a shared connection.
+     *
+     * @event
+     * @param {!string} userID
+     *     A unique value representing this specific user's connection to the
+     *     shared connection. This value is generated by the server and is
+     *     guaranteed to be unique relative to other users of the connection.
+     *
+     * @param {!string} name
+     *     A human-readable name representing the user that joined, such as
+     *     their username. This value is provided by the web application during
+     *     the connection handshake and is not necessarily unique relative to
+     *     other users of the connection.
+     */
+    this.onjoin = null;
+
+    /**
+     * Fired when a user leaves a shared connection.
+     *
+     * @event
+     * @param {!string} userID
+     *     A unique value representing this specific user's connection to the
+     *     shared connection. This value is generated by the server and is
+     *     guaranteed to be unique relative to other users of the connection.
+     *
+     * @param {!string} name
+     *     A human-readable name representing the user that left, such as their
+     *     username. This value is provided by the web application during the
+     *     connection handshake and is not necessarily unique relative to other
+     *     users of the connection.
+     */
+    this.onleave = null;
 
     /**
      * Fired when a audio stream is created. The stream provided to this event
      * handler will contain its own event handlers for received data.
      *
      * @event
-     * @param {Guacamole.InputStream} stream
+     * @param {!Guacamole.InputStream} stream
      *     The stream that will receive audio data from the server.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the audio data which will be received.
      *
      * @return {Guacamole.AudioPlayer}
      *     An object which implements the Guacamole.AudioPlayer interface and
-     *     has been initialied to play the data in the provided stream, or null
+     *     has been initialized to play the data in the provided stream, or null
      *     if the built-in audio players of the Guacamole client should be
      *     used.
      */
@@ -630,38 +770,52 @@ Guacamole.Client = function(tunnel) {
      * handler will contain its own event handlers for received data.
      *
      * @event
-     * @param {Guacamole.InputStream} stream
+     * @param {!Guacamole.InputStream} stream
      *     The stream that will receive video data from the server.
      *
-     * @param {Guacamole.Display.VisibleLayer} layer
+     * @param {!Guacamole.Display.VisibleLayer} layer
      *     The destination layer on which the received video data should be
      *     played. It is the responsibility of the Guacamole.VideoPlayer
      *     implementation to play the received data within this layer.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the video data which will be received.
      *
      * @return {Guacamole.VideoPlayer}
      *     An object which implements the Guacamole.VideoPlayer interface and
-     *     has been initialied to play the data in the provided stream, or null
+     *     has been initialized to play the data in the provided stream, or null
      *     if the built-in video players of the Guacamole client should be
      *     used.
      */
     this.onvideo = null;
 
+    /**
+     * Fired when the remote client is explicitly declaring the level of
+     * multi-touch support provided by a particular display layer.
+     *
+     * @event
+     * @param {!Guacamole.Display.VisibleLayer} layer
+     *     The layer whose multi-touch support level is being declared.
+     *
+     * @param {!number} touches
+     *     The maximum number of simultaneous touches supported by the given
+     *     layer, where 0 indicates that touch events are not supported at all.
+     */
+    this.onmultitouch = null;
+
     /**
      * Fired when the current value of a connection parameter is being exposed
      * by the server.
      *
      * @event
-     * @param {Guacamole.InputStream} stream
+     * @param {!Guacamole.InputStream} stream
      *     The stream that will receive connection parameter data from the
      *     server.
      *
-     * @param {String} mimetype
+     * @param {!string} mimetype
      *     The mimetype of the data which will be received.
      *
-     * @param {String} name
+     * @param {!string} name
      *     The name of the connection parameter whose value is being exposed.
      */
     this.onargv = null;
@@ -670,9 +824,11 @@ Guacamole.Client = function(tunnel) {
      * Fired when the clipboard of the remote client is changing.
      * 
      * @event
-     * @param {Guacamole.InputStream} stream The stream that will receive
-     *                                       clipboard data from the server.
-     * @param {String} mimetype The mimetype of the data which will be received.
+     * @param {!Guacamole.InputStream} stream
+     *     The stream that will receive clipboard data from the server.
+     *
+     * @param {!string} mimetype
+     *     The mimetype of the data which will be received.
      */
     this.onclipboard = null;
 
@@ -681,10 +837,14 @@ Guacamole.Client = function(tunnel) {
      * handler will contain its own event handlers for received data.
      * 
      * @event
-     * @param {Guacamole.InputStream} stream The stream that will receive data
-     *                                       from the server.
-     * @param {String} mimetype The mimetype of the file received.
-     * @param {String} filename The name of the file received.
+     * @param {!Guacamole.InputStream} stream
+     *     The stream that will receive data from the server.
+     *
+     * @param {!string} mimetype
+     *     The mimetype of the file received.
+     *
+     * @param {!string} filename
+     *     The name of the file received.
      */
     this.onfile = null;
 
@@ -694,10 +854,10 @@ Guacamole.Client = function(tunnel) {
      * requesting and handling data.
      *
      * @event
-     * @param {Guacamole.Object} object
+     * @param {!Guacamole.Object} object
      *     The created filesystem object.
      *
-     * @param {String} name
+     * @param {!string} name
      *     The name of the filesystem.
      */
     this.onfilesystem = null;
@@ -707,12 +867,28 @@ Guacamole.Client = function(tunnel) {
      * handler will contain its own event handlers for received data;
      * 
      * @event
-     * @param {Guacamole.InputStream} stream The stream that will receive data
-     *                                       from the server.
-     * @param {String} mimetype The mimetype of the data which will be received.
-     * @param {String} name The name of the pipe.
+     * @param {!Guacamole.InputStream} stream
+     *     The stream that will receive data from the server.
+     *
+     * @param {!string} mimetype
+     *     The mimetype of the data which will be received.
+     *
+     * @param {!string} name
+     *     The name of the pipe.
      */
     this.onpipe = null;
+    
+    /**
+     * Fired when a "required" instruction is received. A required instruction
+     * indicates that additional parameters are required for the connection to
+     * continue, such as user credentials.
+     * 
+     * @event
+     * @param {!string[]} parameters
+     *      The names of the connection parameters that are required to be
+     *      provided for the connection to continue.
+     */
+    this.onrequired = null;
 
     /**
      * Fired whenever a sync instruction is received from the server, indicating
@@ -720,8 +896,8 @@ Guacamole.Client = function(tunnel) {
      * has sent any results.
      * 
      * @event
-     * @param {Number} timestamp The timestamp associated with the sync
-     *                           instruction.
+     * @param {!number} timestamp
+     *     The timestamp associated with the sync instruction.
      */
     this.onsync = null;
 
@@ -731,10 +907,10 @@ Guacamole.Client = function(tunnel) {
      * the default layer, and negative indices refer to buffers.
      *
      * @private
-     * @param {Number} index
+     * @param {!number} index
      *     The index of the layer to retrieve.
      *
-     * @return {Guacamole.Display.VisibleLayer|Guacamole.Layer}
+     * @return {!(Guacamole.Display.VisibleLayer|Guacamole.Layer)}
      *     The layer having the given index.
      */
     var getLayer = function getLayer(index) {
@@ -765,10 +941,10 @@ Guacamole.Client = function(tunnel) {
      * Positive indices refer to visible layers, an index of zero refers to the
      * default layer, and negative indices refer to buffers.
      *
-     * @param {Guacamole.Display.VisibleLayer|Guacamole.Layer} layer
+     * @param {!(Guacamole.Display.VisibleLayer|Guacamole.Layer)} layer
      *     The layer whose index should be determined.
      *
-     * @returns {Number}
+     * @returns {number}
      *     The index of the given layer, or null if no such layer is associated
      *     with this client.
      */
@@ -807,12 +983,22 @@ Guacamole.Client = function(tunnel) {
 
     /**
      * Handlers for all defined layer properties.
+     *
      * @private
+     * @type {!Object.<string, function>}
      */
     var layerPropertyHandlers = {
 
         "miter-limit": function(layer, value) {
             display.setMiterLimit(layer, parseFloat(value));
+        },
+
+        "multi-touch" : function layerSupportsMultiTouch(layer, value) {
+
+            // Process "multi-touch" property only for true visible layers (not off-screen buffers)
+            if (guac_client.onmultitouch && layer instanceof Guacamole.Display.VisibleLayer)
+                guac_client.onmultitouch(layer, parseInt(value));
+
         }
 
     };
@@ -820,7 +1006,9 @@ Guacamole.Client = function(tunnel) {
     /**
      * Handlers for all instruction opcodes receivable by a Guacamole protocol
      * client.
+     *
      * @private
+     * @type {!Object.<string, function>}
      */
     var instructionHandlers = {
 
@@ -1269,6 +1457,44 @@ Guacamole.Client = function(tunnel) {
             }
 
         },
+        
+        "msg" : function(parameters) {
+
+            var userID;
+            var username;
+
+            // Fire general message handling event first
+            var allowDefault = true;
+            var msgid = parseInt(parameters[0]);
+            if (guac_client.onmsg) {
+                allowDefault = guac_client.onmsg(msgid, parameters.slice(1));
+                if (allowDefault === undefined)
+                    allowDefault = true;
+            }
+
+            // Fire message-specific convenience events if not prevented by the
+            // "onmsg" handler
+            if (allowDefault) {
+                switch (msgid) {
+
+                    case Guacamole.Client.Message.USER_JOINED:
+                        userID = parameters[1];
+                        username = parameters[2];
+                        if (guac_client.onjoin)
+                            guac_client.onjoin(userID, username);
+                        break;
+
+                    case Guacamole.Client.Message.USER_LEFT:
+                        userID = parameters[1];
+                        username = parameters[2];
+                        if (guac_client.onleave)
+                            guac_client.onleave(userID, username);
+                        break;
+
+                }
+            }
+            
+        },
 
         "name": function(parameters) {
             if (guac_client.onname) guac_client.onname(parameters[0]);
@@ -1337,6 +1563,10 @@ Guacamole.Client = function(tunnel) {
             display.rect(layer, x, y, w, h);
 
         },
+                
+        "required": function required(parameters) {
+            if (guac_client.onrequired) guac_client.onrequired(parameters);
+        },
         
         "reset": function(parameters) {
 
@@ -1544,9 +1774,12 @@ Guacamole.Client = function(tunnel) {
      * Connects the underlying tunnel of this Guacamole.Client, passing the
      * given arbitrary data to the tunnel during the connection process.
      *
-     * @param data Arbitrary connection data to be sent to the underlying
-     *             tunnel during the connection process.
-     * @throws {Guacamole.Status} If an error occurs during connection.
+     * @param {string} data
+     *     Arbitrary connection data to be sent to the underlying tunnel during
+     *     the connection process.
+     *
+     * @throws {!Guacamole.Status}
+     *     If an error occurs during connection.
      */
     this.connect = function(data) {
 
@@ -1572,7 +1805,9 @@ Guacamole.Client = function(tunnel) {
 
 /**
  * Map of all Guacamole binary raster operations to transfer functions.
+ *
  * @private
+ * @type {!Object.<number, function>}
  */
 Guacamole.Client.DefaultTransferFunction = {
 
@@ -1685,3 +1920,31 @@ Guacamole.Client.DefaultTransferFunction = {
     }
 
 };
+
+/**
+ * A list of possible messages that can be sent by the server for processing
+ * by the client.
+ * 
+ * @type {!Object.<string, number>}
+ */
+Guacamole.Client.Message = {
+    
+    /**
+     * A client message that indicates that a user has joined an existing
+     * connection. This message expects a single additional argument - the
+     * name of the user who has joined the connection.
+     * 
+     * @type {!number}
+     */
+    "USER_JOINED": 0x0001,
+    
+    /**
+     * A client message that indicates that a user has left an existing
+     * connection. This message expects a single additional argument - the
+     * name of the user who has left the connection.
+     * 
+     * @type {!number}
+     */
+    "USER_LEFT": 0x0002
+    
+};
