@@ -67,7 +67,8 @@ public class ExtensionModule extends ServletModule {
             "1.1.0",
             "1.2.0",
             "1.3.0",
-            "1.4.0"
+            "1.4.0",
+            "1.5.0"
         ));
 
     /**
@@ -140,6 +141,13 @@ public class ExtensionModule extends ServletModule {
     private final List<Listener> boundListeners =
             new ArrayList<Listener>();
 
+    /**
+     * All temporary files that should be deleted upon application shutdown, in
+     * reverse order of desired deletion. This will typically simply be the
+     * order that each file was created.
+     */
+    private final List<File> temporaryFiles = new ArrayList<>();
+
     /**
      * Service for adding and retrieving language resources.
      */
@@ -222,7 +230,7 @@ public class ExtensionModule extends ServletModule {
     }
 
     /**
-     * Binds each of the the given AuthenticationProvider classes such that any
+     * Binds each of the given AuthenticationProvider classes such that any
      * service requiring access to the AuthenticationProvider can obtain it via
      * injection.
      *
@@ -261,6 +269,20 @@ public class ExtensionModule extends ServletModule {
         return Collections.unmodifiableList(boundAuthenticationProviders);
     }
 
+    /**
+     * Returns a list of all temporary files that should be deleted upon
+     * application shutdown, in reverse order of desired deletion. This will
+     * typically simply be the order that each file was created.
+     *
+     * @return
+     *     A List of all temporary files that should be deleted upon
+     *     application shutdown. The List is not modifiable.
+     */
+    @Provides
+    public List<File> getTemporaryFiles() {
+        return Collections.unmodifiableList(temporaryFiles);
+    }
+    
     /**
      * Binds the given provider class such that a listener is bound for each
      * listener interface implemented by the provider and such that all bound
@@ -278,7 +300,7 @@ public class ExtensionModule extends ServletModule {
     }
 
     /**
-     * Binds each of the the given Listener classes such that any
+     * Binds each of the given Listener classes such that any
      * service requiring access to the Listener can obtain it via
      * injection.
      *
@@ -479,7 +501,7 @@ public class ExtensionModule extends ServletModule {
             try {
 
                 // Load extension from file
-                Extension extension = new Extension(getParentClassLoader(), extensionFile);
+                Extension extension = new Extension(getParentClassLoader(), extensionFile, temporaryFiles);
 
                 // Validate Guacamole version of extension
                 if (!isCompatible(extension.getGuacamoleVersion())) {
